﻿// ****************************************************************************************
// Copyright (C) 2010, Jorn Skaarud Karlsen 
// All rights reserved. 
//
// This file is part of VoxelloScope (http://www.voxelloscope.org/).
//
// VoxelloScope is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// VoxelloScope is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with VoxelloScope.  If not, see <http://www.gnu.org/licenses/>.
//***************************************************************************************

// -------------------------------------------------------------------------------
// Camera constant buffer - updated often..
// -------------------------------------------------------------------------------

cbuffer Camera
{
   int CameraImagePlanePixelWidth;
   int CameraImagePlanePixelHeight;
   float CameraRayDirectionDelta;
   float3 CameraOrigin;
   float3 CameraImagePlaneOrigin;
   float3 CameraImagePlaneDeltaX;
   float3 CameraImagePlaneDeltaY;
   uint CameraFrameCounter;
}

// -------------------------------------------------------------------------------
// Volume constant buffer - updated seldomly..
// -------------------------------------------------------------------------------

cbuffer Volume
{
   int VolumeLevelMax;
   int VolumeLevelMaxIndex;
   uint VolumePreloadLevel;
   float VolumeOpacityThreshold;
   float2 VolumeIntensityScaleBias;
   uint VolumeBrickCountDim;
   uint VolumeBrickDim;
   uint VolumeBrickBorderSize;
   uint VolumeBrickDimWithBorder;
}

// -------------------------------------------------------------------------------
// Volume textures - updated very seldomly..
// -------------------------------------------------------------------------------

Texture3D<int>     VolumeBricksLevelOctreeTexture;
Texture3D<float>   VolumeBricksMinIntensityOctreeTexture;
Texture3D<float>   VolumeBricksMaxIntensityOctreeTexture;
Texture3D<float>   VolumeBricksDataOctreeTexture;

RWStructuredBuffer<uint> BricksCacheBuffer;
StructuredBuffer<int> CacheMipOffsetBuffer;
StructuredBuffer<uint> VolumeBricksPointerOctreeBuffer;

//Texture3D<float>   VolumeBricksLevelVoxtreeTexture;
//Texture3D<bool>    BricksResidentTexture;
//RWTexture3D<uint>  BricksCacheTexture;

// -------------------------------------------------------------------------------
// Structures..
// -------------------------------------------------------------------------------

struct Ray
{
   float3 Origin;
   float3 Direction;
   float3 Position;
   float2 ScreenCoord;
   float DirectionDelta;
};

struct Pixel
{
   float Intensity;
   float Opacity;
   float Depth;
   float EnterDepth;
   float ExitDepth;
   int Iterations;
};

// -------------------------------------------------------------------------------
// More structures..
// -------------------------------------------------------------------------------

SamplerState VolumeDataTextureSampler 
{
   Filter = MIN_MAG_MIP_LINEAR;
   AddressU = Wrap;
   AddressV = Wrap;
   AddressW = Wrap;
};

struct VS_IN
{
   float4 pos : POSITION;
};

struct PS_IN
{
   float4 pos : SV_POSITION;
};

// -------------------------------------------------------------------------------
// Utility methods..
// -------------------------------------------------------------------------------

// Idea: Precompute intersection planes on ray spawn for all subsequent intersections..
// Note: Assumes the ray direction will NEVER have 0 components..
void IntersectUnitAABB(Ray ray, out float t_in, out float t_out)
{
   float3 bbmax = 0.99999;
   float3 bbmin = 0.00001;

   float3 tmax = (bbmax - ray.Position) / ray.Direction;
   float3 tmin = (bbmin - ray.Position) / ray.Direction;

   t_in = max(0.0, max(max(min(tmax.x, tmin.x), min(tmax.y, tmin.y)), min(tmax.z, tmin.z)));
   t_out = max(0.0, min(min(max(tmax.x, tmin.x), max(tmax.y, tmin.y)), max(tmax.z, tmin.z)));
}

Ray CalculateSubRay(Ray ray, int subDivisions)
{
   float3 scaledOrigin = ray.Origin * subDivisions;
   float3 scaledPosition = ray.Position * subDivisions;
   float3 i = floor(scaledPosition);

   Ray subRay;
   subRay.Origin = scaledOrigin - i;
   subRay.Position = scaledPosition - i;
   subRay.Direction = ray.Direction;
   return subRay;
}

float Sample(float3 brickOffset, float3 position, float lodLevel)
{
   float positionScale = (VolumeBrickDim / (float)VolumeBrickDimWithBorder);
   float3 positionOffset = (VolumeBrickBorderSize / (float)VolumeBrickDimWithBorder);
   
   float3 samplePosition = brickOffset + (position * positionScale + positionOffset) / VolumeBrickCountDim;
   return VolumeBricksDataOctreeTexture.SampleLevel(VolumeDataTextureSampler, samplePosition, lodLevel);
}

float WindowLevel(float intensity)
{
   return saturate(intensity * VolumeIntensityScaleBias.x + VolumeIntensityScaleBias.y);
}

Pixel Combine(Pixel pixel, Ray ray, float contributionDistance, float intensity)
{
   float contribution = (1.0 - exp(-contributionDistance * intensity)) * (1.0 - pixel.Opacity);
   pixel.Intensity += contribution * intensity;
   pixel.Opacity += contribution;
   return pixel;
}

// -------------------------------------------------------------------------------
// Ray marches the voxels in a brick..
// -------------------------------------------------------------------------------

Pixel RayMarchBrickVoxels(Pixel pixel, Ray ray, uint brickPointer, float contributionDistance, float steps, float mipLevel)
{
   float3 brickOffset = float3((brickPointer % VolumeBrickCountDim), 
                     ((brickPointer / VolumeBrickCountDim) % VolumeBrickCountDim), 
                     ((brickPointer / (VolumeBrickCountDim * VolumeBrickCountDim)))) / VolumeBrickCountDim;

   [loop]
   while (steps > 0.0)
   {
      // Todo: New:
      //int voxLevel = VolumeBricksLevelVoxtreeTexture.Load(brickOffset, ray.Position);
	  //int subDivisions = (1 << voxLevel);
      //Ray subRay = CalculateSubRay(ray, subDivisions);
      //IntersectUnitAABB(subRay, t_in, t_out);
      //float subSteps = max(1.0, ceil(t_out - t_in));

	  //if (voxLevel == 255) // Bottom level..
	  //{
         
	  //}
	  // Todo: End

      float intensity = Sample(brickOffset, ray.Position, mipLevel);
      float windowLevel = WindowLevel(intensity);

      pixel = Combine(pixel, ray, contributionDistance, windowLevel);

      if (pixel.Opacity > VolumeOpacityThreshold)
      {
         break;
      }

	  pixel.Iterations++;
      ray.Position += ray.Direction;
      steps--;
   }

   return pixel;
}

// -------------------------------------------------------------------------------
// Makes a Pixel and initializes its values..
// -------------------------------------------------------------------------------

Pixel MakePixel()
{
   Pixel pixel;
   pixel.Intensity = 0.0;
   pixel.Opacity = 0.0;
   pixel.Iterations = 0;
   return pixel;
}

// -------------------------------------------------------------------------------
// Ray marches the scene..
// -------------------------------------------------------------------------------

// Idea: Process the brick table based on window level and thus add larger regions of empty space..
// Idea: By preprocessing, there will also be less stress on the texture sampling since you dont 
// have to sample min and max intensity, just a bool value.. well, also an intensity really.. hmm.
Pixel RayMarch(Ray ray)
{
   Pixel pixel = MakePixel();
   float t_in = 0.0;
   float t_out = 0.0;
   float previousDepth = 0.0; // Todo: Sample from depth-buffer and find start position..

   IntersectUnitAABB(ray, pixel.EnterDepth, pixel.ExitDepth);

   pixel.EnterDepth = max(previousDepth, pixel.EnterDepth);
   pixel.EnterDepth = max(0.0001, pixel.EnterDepth);
   pixel.Depth = pixel.EnterDepth;
   ray.Position += ray.Direction * pixel.EnterDepth;

   [loop]
   while (pixel.Depth < pixel.ExitDepth)
   {
      float3 directionDiff = ray.DirectionDelta * pixel.Depth;
      float lookupLevelFloat = min(VolumeLevelMaxIndex + 0.9999, (0.5 / (directionDiff.x * VolumeBrickDim)));
	  int lookupLevel = (int)lookupLevelFloat;
      int lookupMipLevel = VolumeLevelMaxIndex - lookupLevel;
      int lookupSubDivisions = (1 << lookupLevel);
      int4 lookupLevelPosition = int4((ray.Position * lookupSubDivisions), lookupMipLevel);
      //float fractionMipLevel = 1.0 - (lookupLevelFloat - lookupLevel);
      
	  int levelData = VolumeBricksLevelOctreeTexture.Load(lookupLevelPosition);
	  int level = (levelData & 0x0f);         // 0000 1111
	  int loaded = (levelData & 0x80) >> 7;   // 1000 0000
	  int univalue = (levelData & 0x40) >> 6; // 0100 0000
	  int uvalue = (levelData & 0x20) >> 5;   // 0010 0000

      if (loaded == 0)
      {
	     uint bufferIndex = CacheMipOffsetBuffer[lookupLevel] + lookupLevelPosition.x + lookupLevelPosition.y * lookupSubDivisions + lookupLevelPosition.z * lookupSubDivisions * lookupSubDivisions;
         BricksCacheBuffer[bufferIndex] = 0;
		 level = VolumePreloadLevel;
		 //levelData = VolumeBricksLevelOctreeTexture.Load(lookupLevelPosition);
		 break; // Todo: Remove..
      }

	  int subDivisions = (1 << level);
	  int mipLevel = VolumeLevelMaxIndex - level;
	  int4 levelPosition = int4((ray.Position * subDivisions), mipLevel);

      Ray subRay = CalculateSubRay(ray, subDivisions);
      IntersectUnitAABB(subRay, t_in, t_out);
      float subSteps = max(1.0, ceil(t_out - t_in));

      if (univalue == 1)
      {
         pixel = Combine(pixel, subRay, subSteps, uvalue);
      }
      else
      {
	     uint bufferIndex = CacheMipOffsetBuffer[level] + levelPosition.x + levelPosition.y * subDivisions + levelPosition.z * subDivisions * subDivisions;
         uint brickPointer = VolumeBricksPointerOctreeBuffer[bufferIndex];
		 int contributionDistance = (1 << mipLevel);

         pixel = RayMarchBrickVoxels(pixel, subRay, brickPointer, contributionDistance, subSteps, 0.0);
		 BricksCacheBuffer[bufferIndex] = CameraFrameCounter;
      }

      if (pixel.Opacity > VolumeOpacityThreshold)
      {
         break;
      }

	  pixel.Iterations++;
      pixel.Depth += (subSteps / subDivisions);
      ray.Position = ray.Origin + ray.Direction * pixel.Depth;
   }

   // Debug visualization for number of iterations..
   //pixel.Intensity = pixel.Iterations / 250.0;

   return pixel;
}

// -------------------------------------------------------------------------------
// Main vertex shader method..
// -------------------------------------------------------------------------------

PS_IN VS( VS_IN input )
{
   PS_IN output = (PS_IN)0;
   
   output.pos = input.pos;

   return output;
}

// -------------------------------------------------------------------------------
// Main pixel shader method..
// -------------------------------------------------------------------------------

float4 PS( PS_IN input ) : SV_Target
{
   PS_IN output = (PS_IN)0;

   float x = input.pos.x / (float)CameraImagePlanePixelWidth; // Window relative coordinate.
   float y = input.pos.y / (float)CameraImagePlanePixelHeight; // Window relative coordinate.

   float3 direction = CameraImagePlaneOrigin + (CameraImagePlaneDeltaX * input.pos.x) + (CameraImagePlaneDeltaY * input.pos.y);
   float directionScale = 1.0 / (length(direction) * VolumeBrickDim);

   Ray ray;
   ray.Origin = CameraOrigin;
   ray.Position = CameraOrigin;
   ray.DirectionDelta = CameraRayDirectionDelta * directionScale;
   ray.Direction = direction * directionScale;
   ray.ScreenCoord = input.pos.xy;

   Pixel pixel = RayMarch(ray);
   float4 col = pixel.Intensity;
   col.a = 1.0;
   return col;
}

// -------------------------------------------------------------------------------
// Rendering technique..
// -------------------------------------------------------------------------------

technique11 Render
{
   pass P0
   {
      SetGeometryShader( 0 );
      SetVertexShader( CompileShader( vs_5_0, VS() ) );
      SetPixelShader( CompileShader( ps_5_0, PS() ) );
   }
}
