﻿// ****************************************************************************************
// Copyright (C) 2010, Jorn Skaarud Karlsen 
// All rights reserved. 
//
// This file is part of VoxelloScope (http://www.voxelloscope.org/).
//
// VoxelloScope is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// VoxelloScope is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with VoxelloScope.  If not, see <http://www.gnu.org/licenses/>.
//***************************************************************************************

// This shader voxelizes polygonal meshes using conservative rasterization.

cbuffer ConstantBuffer
{
    row_major matrix World;
	int VolumeDim;
};

struct VS_IN
{
   float4 pos : POSITION;
};

struct GS_IN
{
   float4 pos : POSITION;
};

struct PS_IN
{
   float4 pos : SV_POSITION;
   float4 obj : POSITION;
   int axis : AXIS;
};


RWTexture3D<float> VolumeTexture;
//RWTexture3D<float3> NormalTexture;


// -------------------------------------------------------------------------------
// Main vertex shader method..
// -------------------------------------------------------------------------------

GS_IN VS( VS_IN input )
{
   GS_IN output = (GS_IN)0;

   output.pos = input.pos;

   return output;
}

// -------------------------------------------------------------------------------
// Main geometry shader method..
// -------------------------------------------------------------------------------

[maxvertexcount(3)]
void GS( triangle GS_IN input[3], inout TriangleStream<PS_IN> outputStream )
{
    PS_IN output = (PS_IN)0;

    // Note: Transforms to homogeneous clip coordinates after transforming to camera coordinates.
    // The Rasterizer Stage will use the viewport to transform vertices that are in homogenous clip space 
    // (-1 to 1 for X and Y and 0 to 1 for the Z range) into pixel coordinates.
	// Ref: http://slimdx.org/tutorials/DeviceCreation.php

	// Transform vertices.
	float3 a = mul(input[0].pos, World).xyz;
	float3 b = mul(input[1].pos, World).xyz;
	float3 c = mul(input[2].pos, World).xyz;

	// Calculate face normal.
	float3 normal = cross((b - a), (c - a));
	float3 absNormal = abs(normal);

	// Calculate dominant axis.
	float normalMax = max(absNormal.x, max(absNormal.y, absNormal.z));

	// Rotate based on dominant axis.
	if (normalMax == absNormal.z)
	{
		for (int i = 0; i < 3; i++)
		{
			output.axis = 0;
			output.obj = mul(input[i].pos, World).xyzw * float4(1.0, 1.0, 0.5, 1.0) + float4(0.0, 0.0, 0.5, 0.0);
			output.pos = output.obj;
			outputStream.Append(output);
		}
	}
	else if (normalMax == absNormal.y)
	{
		for (int i = 0; i < 3; i++)
		{
			output.axis = 1;
			output.obj = mul(input[i].pos, World).xzyw  * float4(1.0, 1.0, 0.5, 1.0) + float4(0.0, 0.0, 0.5, 0.0);
			output.pos = output.obj;
			outputStream.Append(output);
		}
	}
	else if (normalMax == absNormal.x)
	{
		for (int i = 0; i < 3; i++)
		{
			output.axis = 2;
			output.obj = mul(input[i].pos, World).zyxw * float4(1.0, 1.0, 0.5, 1.0) + float4(0.0, 0.0, 0.5, 0.0);
			output.pos = output.obj;
			outputStream.Append(output);
		}
	}

	// Todo: Extend polygon for conservative rasterization.
	// Todo: Create bounding box to help discard fragments in pixel shader.
	// Todo: Do ddx and ddy to calculate depth along a polygon. At maximum 3 voxels in depth.
}

// -------------------------------------------------------------------------------
// Main pixel shader method..
// -------------------------------------------------------------------------------

float4 PS( PS_IN input ) : SV_Target
{
   PS_IN output = (PS_IN)0;

   // Note: Homogenous clip space (-1 to 1 for X and Y and 0 to 1 for the Z range).
   // Need to transform from homogeneous clip coordinates to volume coordinates.

   if (input.axis == 0)
   {
     float3 pos = VolumeDim * float3(0.5, 0.5, 1.0) * (input.obj.xyz + float3(1.0, 1.0, 0.0));
     VolumeTexture[pos] = 1.0;
     return float4(1.0, 0.0, 0.0, 1.0);
   }
   else if (input.axis == 1)
   {
     float3 pos = VolumeDim * float3(0.5, 1.0, 0.5) * (input.obj.xzy + float3(1.0, 0.0, 1.0));
     VolumeTexture[pos] = 1.0;
     return float4(0.0, 1.0, 0.0, 1.0);
   }
   else if (input.axis == 2)
   {
     float3 pos = VolumeDim * float3(1.0, 0.5, 0.5) * (input.obj.zyx + float3(0.0, 1.0, 1.0));
     VolumeTexture[pos] = 1.0;
     return float4(1.0, 0.0, 1.0, 1.0);
   }

   //discard;
   return float4(0.5, 1.0, 0.5, 1.0);
}

// -------------------------------------------------------------------------------
// Rendering technique..
// -------------------------------------------------------------------------------

technique11 Render
{
   pass P0
   {
      SetGeometryShader( CompileShader( gs_5_0, GS() ) );
      SetVertexShader( CompileShader( vs_5_0, VS() ) );
      SetPixelShader( CompileShader( ps_5_0, PS() ) );
   }
}
